/* $Id: inclued.c,v 1.9 2007/06/13 21:55:43 gopalv Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "inclued.h"

/* {{{ stupid stringifcation */
#if DEFAULT_SLASH == '/'
	#define DEFAULT_SLASH_STRING "/"
#elif DEFAULT_SLASH == '\\'
	#define DEFAULT_SLASH_STRING "\\"
#else
	#error "Unknown value for DEFAULT_SLASH"
#endif
/* }}} */

/* {{{ inclued globals 
 *
 * true globals, no need for thread safety here 
 */
/* nothing for now */
/* }}} */

/* {{{ ZEND_DECLARE_MODULE_GLOBALS(inclued) */
ZEND_DECLARE_MODULE_GLOBALS(inclued)
/* }}} */

/* {{{ php_inclued_init_globals */
static void php_inclued_init_globals(zend_inclued_globals* inclued_globals TSRMLS_DC)
{
	inclued_globals->enabled = 0;
	inclued_globals->counter = 0;
	inclued_globals->dumpdir = 0;
	inclued_globals->includes_hash = NULL;
}
/* }}} */

/* {{{ php_inclued_shutdown_globals */
static void php_inclued_shutdown_globals(zend_inclued_globals* inclued_globals TSRMLS_DC)
{
	/* nothing ? */
}
/* }}} */

/* {{{ ini entries */
PHP_INI_BEGIN()
	STD_PHP_INI_BOOLEAN("inclued.enabled", "0",  PHP_INI_SYSTEM, OnUpdateBool,  enabled,   zend_inclued_globals, inclued_globals)
	STD_PHP_INI_BOOLEAN("inclued.dumpdir", NULL,  PHP_INI_SYSTEM, OnUpdateString,  dumpdir,   zend_inclued_globals, inclued_globals)
PHP_INI_END()
/* }}} */

/* {{{ PHP_FUNCTION declarations */
PHP_FUNCTION(inclued_get_data);
/* }}} */

/* {{{ inclued_functions[]
 *
 * Every user visible function must have an entry in inclued_functions[].
 */
zend_function_entry inclued_functions[] = {
	PHP_FE(inclued_get_data,		NULL)
	{NULL, NULL, NULL}	/* Must be the last line in inclued_functions[] */
};
/* }}} */

/* {{{ inclued_module_entry
 */
zend_module_entry inclued_module_entry = {
	STANDARD_MODULE_HEADER,
	"inclued",
	inclued_functions,
	PHP_MINIT(inclued),
	PHP_MSHUTDOWN(inclued),
	PHP_RINIT(inclued),
	PHP_RSHUTDOWN(inclued),
	PHP_MINFO(inclued),
#if ZEND_MODULE_API_NO >= 20010901
	"@PACKAGE_VERSION@",
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_INCLUED
ZEND_GET_MODULE(inclued)
#endif

/* {{{ pack_output */
void pack_output(zval* return_value)
{

	zval **data;
	zval *includes_hash;
	zval *request_hash;
	
	MAKE_STD_ZVAL(includes_hash);

	*includes_hash = *INCLUED_G(includes_hash);

	zval_copy_ctor(includes_hash);

	MAKE_STD_ZVAL(request_hash);
	
	array_init(request_hash);
	
#define PULL_ZVAL(symtab, name)  do { \
		if (zend_hash_find((symtab), #name , sizeof(#name), (void **) &data) != FAILURE) \
		{ \
			zval *copy_zval;\
			MAKE_STD_ZVAL(copy_zval);\
			*copy_zval = *data[0];\
			zval_copy_ctor(copy_zval);\
			add_assoc_zval(request_hash, #name, copy_zval);\
		}\
	} while(0);

	PULL_ZVAL(&EG(symbol_table),PHP_SELF);
	PULL_ZVAL(&EG(symbol_table),_REQUEST);
	PULL_ZVAL(&EG(symbol_table),_COOKIE);

	if (zend_hash_find((&EG(symbol_table)), "_SERVER", sizeof("_SERVER"), (void **) &data) != FAILURE) 
	{
		HashTable *_server = Z_ARRVAL_PP(data);
		PULL_ZVAL(_server, SCRIPT_FILENAME);
		PULL_ZVAL(_server, REQUEST_URI);
		PULL_ZVAL(_server, REQUEST_TIME);
	}

	array_init(return_value);
	
	add_assoc_zval(return_value, "request", request_hash);
	add_assoc_zval(return_value, "includes", includes_hash);
	
	return;
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(inclued)
{
	ZEND_INIT_MODULE_GLOBALS(inclued, php_inclued_init_globals, php_inclued_shutdown_globals);

	REGISTER_INI_ENTRIES();
	
	if(INCLUED_G(enabled)) 
	{
		inclued_zend_init(TSRMLS_C);
	}
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 *  */
PHP_MSHUTDOWN_FUNCTION(inclued)
{

	if(INCLUED_G(enabled)) 
	{
		inclued_zend_shutdown(TSRMLS_C);
	}
	
#ifdef ZTS
	ts_free_id(inclued_globals_id);
#else
	php_inclued_shutdown_globals(&inclued_globals);
#endif
	UNREGISTER_INI_ENTRIES();
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_RINIT_FUNCTION(inclued) */
PHP_RINIT_FUNCTION(inclued)
{
	zval **data = NULL;
	
    if(INCLUED_G(enabled)) {
		/* allocate new info table */
		ALLOC_INIT_ZVAL(INCLUED_G(includes_hash));
		array_init(INCLUED_G(includes_hash));

		INCLUED_G(counter)++;
    }
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_RSHUTDOWN_FUNCTION(inclued) */
PHP_RSHUTDOWN_FUNCTION(inclued)
{
	char path[MAXPATHLEN];
	FILE *dumpfile;
	smart_str buf = {0};
	php_serialize_data_t var_hash;
	zval retval;
	zval *pretval = &retval;
	
	if(INCLUED_G(enabled) && INCLUED_G(dumpdir)) {
		snprintf(path, MAXPATHLEN, "%s" DEFAULT_SLASH_STRING "inclued.%05d.%d", 
					INCLUED_G(dumpdir), 
					getpid(), INCLUED_G(counter));

		if(!(dumpfile = fopen(path, "w"))) {
			zend_error(E_WARNING, "cannot write to %s" DEFAULT_SLASH_STRING, 
						INCLUED_G(dumpdir));
		}

		pack_output(pretval);

		PHP_VAR_SERIALIZE_INIT(var_hash);
		php_var_serialize(&buf, (zval**)&(pretval), &var_hash TSRMLS_CC);
		PHP_VAR_SERIALIZE_DESTROY(var_hash);
		
		fwrite(buf.c, buf.len, 1, dumpfile);
		fclose(dumpfile);
		smart_str_free(&buf);
		zval_dtor(pretval);
	}

    if(INCLUED_G(enabled)) {
		/* free allocated stuff */
		zval_dtor(INCLUED_G(includes_hash));
		INCLUED_G(includes_hash) = NULL;
    }
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(inclued)
{
	const char * ini_path = NULL;
	int module_number = zend_module->module_number;
	
	php_info_print_table_start();
	php_info_print_table_row(2, "inclued support", "enabled");
	php_info_print_table_end();
}
/* }}} */

/* {{{ proto mixed inclued_get_data(void)
 */
PHP_FUNCTION(inclued_get_data)
{
	pack_output(return_value);
	return;
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
