/* $Id: inclued_zend.c,v 1.9 2007/08/08 01:14:06 gopalv Exp $ */

#include "inclued.h"
#include "inclued_zend.h"

/* Originally taken from Zend/zend_vm_execute.h */

#define _CONST_CODE  0
#define _TMP_CODE    1
#define _VAR_CODE    2
#define _UNUSED_CODE 3
#define _CV_CODE     4
static inline int _inclued_opcode_handler_decode(zend_op *opline)
{
	static const int inclued_vm_decode[] = {
		_UNUSED_CODE, /* 0              */
		_CONST_CODE,  /* 1 = IS_CONST   */
		_TMP_CODE,    /* 2 = IS_TMP_VAR */
		_UNUSED_CODE, /* 3              */
		_VAR_CODE,    /* 4 = IS_VAR     */
		_UNUSED_CODE, /* 5              */
		_UNUSED_CODE, /* 6              */
		_UNUSED_CODE, /* 7              */
		_UNUSED_CODE, /* 8 = IS_UNUSED  */
		_UNUSED_CODE, /* 9              */
		_UNUSED_CODE, /* 10             */
		_UNUSED_CODE, /* 11             */
		_UNUSED_CODE, /* 12             */
		_UNUSED_CODE, /* 13             */
		_UNUSED_CODE, /* 14             */
		_UNUSED_CODE, /* 15             */
		_CV_CODE      /* 16 = IS_CV     */
	};
	return (opline->opcode * 25) + (inclued_vm_decode[opline->op1.op_type] * 5) + inclued_vm_decode[opline->op2.op_type];
}

# define INCLUED_OPCODE_HANDLER_DECODE(opline)	_inclued_opcode_handler_decode(opline)
# if PHP_MAJOR_VERSION >= 6
#  define INCLUED_OPCODE_HANDLER_COUNT			((25 * 152) + 1)
# else
#  define INCLUED_OPCODE_HANDLER_COUNT			((25 * 151) + 1)
# endif
# define INCLUED_REPLACE_OPCODE(opname)         { int i; for(i = 0; i < 25; i++) if (zend_opcode_handlers[(opname*25) + i]) zend_opcode_handlers[(opname*25) + i] = inclued_op_##opname; }

static opcode_handler_t *inclued_original_opcode_handlers;
static opcode_handler_t inclued_opcode_handlers[INCLUED_OPCODE_HANDLER_COUNT];

#define INCLUED_EX_T(offset)					(*(temp_variable *)((char*)execute_data->Ts + offset))

static zval *inclued_get_zval_ptr(znode *node, zval **freeval, zend_execute_data *execute_data TSRMLS_DC)
{
	*freeval = NULL;

	switch (node->op_type) {
		case IS_CONST:
			return &(node->u.constant);
		case IS_VAR:
			return INCLUED_EX_T(node->u.var).var.ptr;
		case IS_TMP_VAR:
			return (*freeval = &INCLUED_EX_T(node->u.var).tmp_var);
#ifdef ZEND_ENGINE_2_1
		case IS_CV:
		{
			zval ***ret = &execute_data->CVs[node->u.var];

			if (!*ret) {
				zend_compiled_variable *cv = &EG(active_op_array)->vars[node->u.var];

				if (zend_hash_quick_find(EG(active_symbol_table), cv->name, cv->name_len+1, cv->hash_value, (void**)ret)==FAILURE) {
					zend_error(E_NOTICE, "Undefined variable: %s", cv->name);
					return &EG(uninitialized_zval);
				}
			}
			return **ret;
		}
#endif
		case IS_UNUSED:
		default:
			return NULL;
	}
}

static int add_include(ZEND_OPCODE_HANDLER_ARGS)
{
	zend_op *opline = execute_data->opline;
	zval * includes_hash = INCLUED_G(includes_hash);
	char * operation;

	zval * include_entry;

	php_stream_wrapper *wrapper;
	zval *inc_filename = NULL, tmp_inc_filename;
	zval *freeop1 = NULL;
	char *path_for_open;
	zend_file_handle file_handle;
	int op_type = opline->op2.u.constant.value.lval;


	assert(includes_hash != NULL);

	switch(op_type)
	{
		case ZEND_INCLUDE:
		{
			operation = "include";
		}
		break;
		case ZEND_REQUIRE:
		{
			operation = "require";
		}
		break;
		case ZEND_REQUIRE_ONCE:
		{
			operation = "require_once";
		}
		break;
		case ZEND_INCLUDE_ONCE:
		{
			operation = "include_once";
		}
		break;
		case ZEND_EVAL:
		{
			operation = "eval";
		}
		break;
	}
	
	if(op_type == ZEND_EVAL)
	{
		MAKE_STD_ZVAL(include_entry);
		
		array_init(include_entry);
		add_assoc_string(include_entry, "operation", operation, 1);
		add_assoc_long(include_entry, "op_type", op_type);
		add_assoc_string(include_entry, "fromfile",	execute_data->op_array->filename, 1);
		add_assoc_long(include_entry, "fromline", opline->lineno);

		if(execute_data->op_array->function_name) 
		{
			add_assoc_string(include_entry, "function", (char*)execute_data->op_array->function_name, 1);
		}

		add_next_index_zval(includes_hash, include_entry);

		return;
	}

	inc_filename = inclued_get_zval_ptr(&opline->op1, &freeop1, execute_data TSRMLS_CC);

	if(Z_TYPE_P(inc_filename) != IS_STRING) 
	{
		tmp_inc_filename = *inc_filename;
		zval_copy_ctor(&tmp_inc_filename);
		convert_to_string(&tmp_inc_filename);
		inc_filename = &tmp_inc_filename;
	}

	wrapper = php_stream_locate_url_wrapper(Z_STRVAL_P(inc_filename), &path_for_open, 0 TSRMLS_CC);


	if(SUCCESS == zend_stream_open(inc_filename->value.str.val, &file_handle TSRMLS_CC))
	{
		/* sledgehammer ! */
		if (!file_handle.opened_path)
		{
			file_handle.opened_path = estrndup(inc_filename->value.str.val, inc_filename->value.str.len);
		}

		/*
		fprintf(stderr, "%s%s('%s') from %s:%d\n", 
			(op_type & (ZEND_INCLUDE_ONCE | ZEND_INCLUDE)) ? "include" : "require",
			(op_type & (ZEND_INCLUDE_ONCE | ZEND_REQUIRE_ONCE)) ? "_once" : "",
			file_handle.opened_path,
			execute_data->op_array->filename,
			opline->lineno);
		*/

		MAKE_STD_ZVAL(include_entry);
		
		array_init(include_entry);

		add_assoc_string(include_entry, "operation", operation, 1);
		add_assoc_long(include_entry, "op_type", op_type);
		add_assoc_string(include_entry, "filename", Z_STRVAL_P(inc_filename), 1);
		add_assoc_string(include_entry, "opened_path", file_handle.opened_path, 1);

		if(zend_hash_exists(&EG(included_files), file_handle.opened_path, strlen(file_handle.opened_path)+1))
		{
			add_assoc_bool(include_entry, "duplicate", 1);
		}
		
		add_assoc_string(include_entry, "fromfile",	execute_data->op_array->filename, 1);
		add_assoc_long(include_entry, "fromline", opline->lineno);

		if(execute_data->op_array->function_name) 
		{
			add_assoc_string(include_entry, "function", (char*)execute_data->op_array->function_name, 1);
		}

		if(opline->op1.op_type != IS_CONST)
		{
			add_assoc_bool(include_entry, "variable_include", 1);
		}
		
		if(wrapper != &php_plain_files_wrapper)
		{
			add_assoc_string(include_entry, "streamwrapper", (char*)wrapper->wops->label, 1);
		}

		add_next_index_zval(includes_hash, include_entry);

		zend_file_handle_dtor(&file_handle);
	}

	if(inc_filename==&tmp_inc_filename)
	{
		zval_dtor(&tmp_inc_filename);
	}
}

static int inclued_op_ZEND_INCLUDE_OR_EVAL(ZEND_OPCODE_HANDLER_ARGS)
{
	zend_op *opline = execute_data->opline;

	add_include(ZEND_OPCODE_HANDLER_ARGS_PASSTHRU);
	return inclued_original_opcode_handlers[INCLUED_OPCODE_HANDLER_DECODE(opline)](ZEND_OPCODE_HANDLER_ARGS_PASSTHRU);
}


void inclued_zend_init(TSRMLS_D)
{
	if(INCLUED_G(enabled))
	{
		/* replace opcode handler array */
		memcpy(inclued_opcode_handlers, zend_opcode_handlers, sizeof(inclued_opcode_handlers));
		inclued_original_opcode_handlers = zend_opcode_handlers;
		zend_opcode_handlers = inclued_opcode_handlers;

		INCLUED_REPLACE_OPCODE(ZEND_INCLUDE_OR_EVAL);
	}
}

void inclued_zend_shutdown(TSRMLS_D)
{
	if(INCLUED_G(enabled))
	{
		zend_opcode_handlers = inclued_original_opcode_handlers;
	}
}
